use crate::{
    formats::PixelFormatSpec,
    math::{self, PixelBox, PixelSize},
};

pub trait BitmapData {
    fn data(&self) -> &[u8];
    fn width(&self) -> usize;
    fn height(&self) -> usize;
    fn stride(&self) -> usize;

    fn size(&self) -> PixelSize {
        PixelSize::new(self.width() as i32, self.height() as i32)
    }
}

pub trait BitmapDataMut: BitmapData {
    fn data_mut(&mut self) -> &mut [u8];
}

pub trait MakeMutView {
    fn make_view_mut<Format: PixelFormatSpec>(&mut self, area: PixelBox) -> BitmapMut<Format>;
}

impl<'l> MakeMutView for &'l mut dyn BitmapDataMut {
    fn make_view_mut<Format: PixelFormatSpec>(&mut self, area: PixelBox) -> BitmapMut<Format> {
        let stride = self.stride();
        let data_offset =
            (area.min.y as usize * stride) + area.min.x as usize * Format::PIXEL_WIDTH;
        BitmapMut::new_with_stride(
            &mut self.data_mut()[data_offset..],
            area.width() as usize,
            area.height() as usize,
            stride,
        )
    }
}

#[derive(PartialEq, Clone)]
pub struct OwnedBitmapData {
    pub(crate) data: Vec<u8>,
    pub(crate) width: usize,
    pub(crate) height: usize,
    pub(crate) stride: usize,
}

impl BitmapData for OwnedBitmapData {
    fn data(&self) -> &[u8] {
        self.data.as_slice()
    }
    fn width(&self) -> usize {
        self.width
    }
    fn height(&self) -> usize {
        self.height
    }
    fn stride(&self) -> usize {
        self.stride
    }
}

impl BitmapDataMut for OwnedBitmapData {
    fn data_mut(&mut self) -> &mut [u8] {
        self.data.as_mut_slice()
    }
}

#[derive(Clone)]
pub struct BorrowedBitmapData<'l> {
    pub(crate) data: &'l [u8],
    pub(crate) width: usize,
    pub(crate) height: usize,
    pub(crate) stride: usize,
}

impl<'l> BitmapData for BorrowedBitmapData<'l> {
    fn data(&self) -> &[u8] {
        self.data
    }
    fn width(&self) -> usize {
        self.width
    }
    fn height(&self) -> usize {
        self.height
    }
    fn stride(&self) -> usize {
        self.stride
    }
}

pub struct MutableBitmapData<'l> {
    pub(crate) data: &'l mut [u8],
    pub(crate) width: usize,
    pub(crate) height: usize,
    pub(crate) stride: usize,
}

impl<'l> BitmapData for MutableBitmapData<'l> {
    fn data(&self) -> &[u8] {
        self.data
    }
    fn width(&self) -> usize {
        self.width
    }
    fn height(&self) -> usize {
        self.height
    }
    fn stride(&self) -> usize {
        self.stride
    }
}

impl<'l> BitmapDataMut for MutableBitmapData<'l> {
    fn data_mut(&mut self) -> &mut [u8] {
        self.data
    }
}

pub trait BitmapAccess<Format: PixelFormatSpec> {
    fn data(&self) -> &dyn BitmapData;
    fn width(&self) -> usize;
    fn height(&self) -> usize;

    fn size(&self) -> math::PixelSize {
        math::PixelSize::new(self.width() as i32, self.height() as i32)
    }

    fn row_stride(&self) -> usize;

    fn make_view(&self, area: PixelBox) -> BitmapRef<Format> {
        let data_offset = (area.min.y as usize * self.data().stride())
            + area.min.x as usize * Format::PIXEL_WIDTH;
        BitmapRef::new_with_stride(
            &self.data().data()[data_offset..],
            area.width() as usize,
            area.height() as usize,
            self.data().stride(),
        )
    }
}

pub trait BitmapMutAccess<Format: PixelFormatSpec>: BitmapAccess<Format> {
    fn data_mut(&mut self) -> &mut dyn BitmapDataMut;
}

/// Owned bitmap with static pixel format information.
#[derive(Clone, PartialEq)]
pub struct Bitmap<Format: PixelFormatSpec> {
    data: OwnedBitmapData,
    _ghost: std::marker::PhantomData<Format>,
}

impl<Format: PixelFormatSpec> Bitmap<Format> {
    pub fn new(width: usize, height: usize) -> Self {
        Self {
            data: OwnedBitmapData {
                data: vec![0; width * height * Format::PIXEL_WIDTH],
                width,
                height,
                stride: width * Format::PIXEL_WIDTH,
            },
            _ghost: Default::default(),
        }
    }

    pub fn new_from_vec(data: Vec<u8>, width: usize, height: usize, stride: usize) -> Self {
        Self {
            data: OwnedBitmapData {
                data,
                width,
                height,
                stride,
            },
            _ghost: Default::default(),
        }
    }

    pub fn as_ref(&self) -> BitmapRef<Format> {
        BitmapRef::new_with_stride(
            &self.data.data,
            self.data.width,
            self.data.height,
            self.data.stride,
        )
    }

    pub fn as_mut(&mut self) -> BitmapMut<Format> {
        BitmapMut::new_with_stride(
            &mut self.data.data,
            self.data.width,
            self.data.height,
            self.data.stride,
        )
    }
}

impl<Format: PixelFormatSpec> BitmapAccess<Format> for Bitmap<Format> {
    fn data(&self) -> &dyn BitmapData {
        &self.data
    }
    fn width(&self) -> usize {
        self.data.width
    }
    fn height(&self) -> usize {
        self.data.height
    }
    fn row_stride(&self) -> usize {
        self.data.stride
    }
}

impl<'l, Format: PixelFormatSpec, T: BitmapAccess<Format>> BitmapAccess<Format> for &'l T {
    fn data(&self) -> &dyn BitmapData {
        (*self).data()
    }
    fn size(&self) -> math::PixelSize {
        (*self).size()
    }
    fn width(&self) -> usize {
        (*self).width()
    }
    fn height(&self) -> usize {
        (*self).height()
    }
    fn row_stride(&self) -> usize {
        (*self).row_stride()
    }
    fn make_view(&self, area: PixelBox) -> BitmapRef<Format> {
        (*self).make_view(area)
    }
}

impl<Format: PixelFormatSpec> BitmapMutAccess<Format> for Bitmap<Format> {
    fn data_mut(&mut self) -> &mut dyn BitmapDataMut {
        &mut self.data
    }
}

impl<Format: PixelFormatSpec> std::fmt::Debug for Bitmap<Format> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("Bitmap")
            .field("width", &self.data.width)
            .field("height", &self.data.height)
            .field("stride", &self.data.stride)
            .finish_non_exhaustive()
    }
}

/// Borrowed bitmap with static pixel format information.
#[derive(Clone)]
pub struct BitmapRef<'l, Format: PixelFormatSpec> {
    data: BorrowedBitmapData<'l>,
    _ghost: std::marker::PhantomData<Format>,
}

impl<'l, Format: PixelFormatSpec> BitmapRef<'l, Format> {
    /// Create a new bitmap reference using the default row stride.
    pub fn new(data: &'l [u8], width: usize, height: usize) -> Self {
        Self {
            data: BorrowedBitmapData {
                data,
                width,
                height,
                stride: width * Format::PIXEL_WIDTH,
            },
            _ghost: Default::default(),
        }
    }

    /// Create a new bitmap reference using a specific row stride.
    pub fn new_with_stride(data: &'l [u8], width: usize, height: usize, stride: usize) -> Self {
        Self {
            data: BorrowedBitmapData {
                data,
                width,
                height,
                stride,
            },
            _ghost: Default::default(),
        }
    }
}

impl<'l, Format: PixelFormatSpec> BitmapAccess<Format> for BitmapRef<'l, Format> {
    fn data(&self) -> &dyn BitmapData {
        &self.data
    }
    fn width(&self) -> usize {
        self.data.width
    }
    fn height(&self) -> usize {
        self.data.height
    }
    fn row_stride(&self) -> usize {
        self.data.stride
    }
}

/// Mutably borrowed bitmap with static pixel format information.
pub struct BitmapMut<'l, Format: PixelFormatSpec> {
    data: MutableBitmapData<'l>,
    _ghost: std::marker::PhantomData<Format>,
}

impl<'l, Format: PixelFormatSpec> BitmapMut<'l, Format> {
    /// Create a new mutable bitmap reference using the default row stride.
    pub fn new(data: &'l mut [u8], width: usize, height: usize) -> Self {
        Self {
            data: MutableBitmapData {
                data,
                width,
                height,
                stride: width * Format::PIXEL_WIDTH,
            },
            _ghost: Default::default(),
        }
    }

    /// Create a new mutable bitmap reference using a specific row stride.
    pub fn new_with_stride(data: &'l mut [u8], width: usize, height: usize, stride: usize) -> Self {
        Self {
            data: MutableBitmapData {
                data,
                width,
                height,
                stride,
            },
            _ghost: Default::default(),
        }
    }
}

impl<'l, Format: PixelFormatSpec> BitmapAccess<Format> for BitmapMut<'l, Format> {
    fn data(&self) -> &dyn BitmapData {
        &self.data
    }
    fn width(&self) -> usize {
        self.data.width
    }
    fn height(&self) -> usize {
        self.data.height
    }
    fn row_stride(&self) -> usize {
        self.data.stride
    }
}

impl<'l, Format: PixelFormatSpec> BitmapMutAccess<Format> for BitmapMut<'l, Format> {
    fn data_mut(&mut self) -> &mut dyn BitmapDataMut {
        &mut self.data
    }
}
