use crate::formats::{PackedLocation, PixelFormatSpec};

/// Helper type alias for Americans.
pub type Color = Colour;

/// Colour specification type.
///
/// This type internally stores the colour as RGBA, and each pixel format will use whatever subset
/// of the colour values are relevant. For example, an `A8` surface will ignore the R/G/B channels.
/// 'Missing' channels will be given a default value of 255.
#[derive(Clone, Copy, Debug, PartialEq)]
pub struct Colour([u8; 4]);

/// Basic colour creation and manipulation functions
impl Colour {
    pub const fn rgba(r: u8, g: u8, b: u8, a: u8) -> Self {
        Self([r, g, b, a])
    }

    pub const fn rgb(r: u8, g: u8, b: u8) -> Self {
        Self([r, g, b, 255])
    }

    pub fn hex_rgb(hex: &str) -> Option<Self> {
        // acceptable lengths: 6, 7
        let toparse = match (hex.len(), hex.chars().next()) {
            (6, _) => hex,
            (7, Some('#')) => &hex[1..],
            _ => return None,
        };

        Some(Self([
            u8::from_str_radix(&toparse[0..2], 16).ok()?,
            u8::from_str_radix(&toparse[2..4], 16).ok()?,
            u8::from_str_radix(&toparse[4..6], 16).ok()?,
            255,
        ]))
    }

    pub const fn r(&self) -> u8 {
        self.0[0]
    }
    pub const fn with_r(mut self, r: u8) -> Self {
        self.0[0] = r;
        self
    }

    pub const fn g(&self) -> u8 {
        self.0[1]
    }
    pub const fn with_g(mut self, g: u8) -> Self {
        self.0[1] = g;
        self
    }

    pub const fn b(&self) -> u8 {
        self.0[2]
    }
    pub const fn with_b(mut self, b: u8) -> Self {
        self.0[2] = b;
        self
    }

    pub const fn a(&self) -> u8 {
        self.0[3]
    }
    pub const fn with_a(mut self, a: u8) -> Self {
        self.0[3] = a;
        self
    }
}

/// Generic conversion functions
impl Colour {
    #[inline]
    fn convert_pack(
        dst: &mut [u8],
        src: &[u8],
        src_pack: PackedLocation,
        dst_pack: PackedLocation,
    ) {
        match (src_pack.index(), dst_pack.index()) {
            (Some(srcindex), Some(dstindex)) => {
                dst[dstindex as usize] = src[srcindex as usize];
            }
            (None, Some(dstindex)) => {
                dst[dstindex as usize] = 255;
            }
            (_, None) => (),
        }
    }

    pub(crate) fn convert_byte_form<SrcFormat: PixelFormatSpec, DstFormat: PixelFormatSpec>(
        dst: &mut [u8],
        src: &[u8],
    ) {
        Self::convert_pack(dst, src, SrcFormat::RED_PACK, DstFormat::RED_PACK);
        Self::convert_pack(dst, src, SrcFormat::GREEN_PACK, DstFormat::GREEN_PACK);
        Self::convert_pack(dst, src, SrcFormat::BLUE_PACK, DstFormat::BLUE_PACK);
        Self::convert_pack(dst, src, SrcFormat::ALPHA_PACK, DstFormat::ALPHA_PACK);
    }

    pub(crate) fn write_as_bytes<Format: PixelFormatSpec>(&self, to: &mut [u8]) {
        if let Some(offset) = Format::RED_PACK.index() {
            to[offset as usize] = self.r();
        }
        if let Some(offset) = Format::GREEN_PACK.index() {
            to[offset as usize] = self.g();
        }
        if let Some(offset) = Format::BLUE_PACK.index() {
            to[offset as usize] = self.b();
        }
        if let Some(offset) = Format::ALPHA_PACK.index() {
            to[offset as usize] = self.a();
        }
    }

    pub(crate) fn read_from_bytes<Format: PixelFormatSpec>(from: &[u8]) -> Self {
        Self::rgba(
            if let Some(offset) = Format::RED_PACK.index() {
                from[offset as usize]
            } else {
                255
            },
            if let Some(offset) = Format::GREEN_PACK.index() {
                from[offset as usize]
            } else {
                255
            },
            if let Some(offset) = Format::BLUE_PACK.index() {
                from[offset as usize]
            } else {
                255
            },
            if let Some(offset) = Format::ALPHA_PACK.index() {
                from[offset as usize]
            } else {
                255
            },
        )
    }
}

/// Colour constants
impl Colour {
    pub const BLACK: Colour = Self::rgba(0, 0, 0, 255);
    pub const WHITE: Colour = Self::rgba(255, 255, 255, 255);
    pub const GREY: Colour = Self::rgba(128, 128, 128, 255);
    pub const TRANSPARENT_GREY: Colour = Self::rgba(128, 128, 128, 128);
    pub const RED: Colour = Self::rgba(255, 0, 0, 255);
    pub const DARK_RED: Colour = Self::rgba(64, 0, 0, 255);
    pub const GREEN: Colour = Self::rgba(0, 255, 0, 255);
    pub const DARK_GREEN: Colour = Self::rgba(0, 64, 0, 255);
    pub const BLUE: Colour = Self::rgba(0, 0, 255, 255);
    pub const DARK_BLUE: Colour = Self::rgba(0, 0, 64, 255);
}

#[derive(Clone, Copy, PartialEq, Debug)]
/// Describes a blending mode between a backdrop and a source pixel.
pub enum BlendMode {
    /// Use source colour without doing any blending.
    SourceCopy,
    /// Use the alpha value of both the source and backdrop pixels.
    Simple,
    /// Only use the alpha value of the source.
    SourceAlpha,
    /// Only use the alpha value of the backdrop.
    BackdropAlpha,
    /// Multiply colour channels together.
    Multiply,
}

impl BlendMode {
    // computes a*alpha + b*(1-alpha)
    const fn mix(a: u8, b: u8, alpha: u8) -> u8 {
        (((a as u16 * alpha as u16) + (b as u16 * (255 - alpha as u16))) / 255) as u8
    }

    // computes a*alpha1 + b*alpha2*(1-alpha)
    const fn mix2(a: u8, b: u8, alpha1: u8, alpha2: u8) -> u8 {
        let a_term = a as u32 * alpha1 as u32 * 255;
        let b_term = b as u32 * alpha2 as u32 * (255 - alpha1) as u32;
        ((a_term + b_term) / (255 * 255)) as u8
    }

    pub const fn blend(&self, source: &Colour, backdrop: &Colour) -> Colour {
        // Alpha compositing colour and alpha formulae: [1]
        // - co = Cs x αs + Cb x αb x (1 - αs)
        // - αo = αs + αb x (1 - αs)
        // [1]: https://www.w3.org/TR/compositing/#simplealphacompositing
        match self {
            Self::SourceCopy => *source,
            Self::Simple => Colour::rgba(
                Self::mix2(source.r(), backdrop.r(), source.a(), backdrop.a()),
                Self::mix2(source.g(), backdrop.g(), source.a(), backdrop.a()),
                Self::mix2(source.b(), backdrop.b(), source.a(), backdrop.a()),
                Self::mix(255, backdrop.a(), source.a()),
            ),
            Self::SourceAlpha => Colour::rgba(
                Self::mix(source.r(), backdrop.r(), source.a()),
                Self::mix(source.g(), backdrop.g(), source.a()),
                Self::mix(source.b(), backdrop.b(), source.a()),
                Self::mix(255, backdrop.a(), source.a()),
            ),
            Self::BackdropAlpha => Colour::rgba(
                Self::mix(source.r(), backdrop.r(), backdrop.a()),
                Self::mix(source.g(), backdrop.g(), backdrop.a()),
                Self::mix(source.b(), backdrop.b(), backdrop.a()),
                Self::mix(255, source.a(), backdrop.a()),
            ),
            Self::Multiply => Colour::rgba(
                ((source.r() as u16 * backdrop.r() as u16) / 255) as u8,
                ((source.g() as u16 * backdrop.g() as u16) / 255) as u8,
                ((source.b() as u16 * backdrop.b() as u16) / 255) as u8,
                ((source.a() as u16 * backdrop.a() as u16) / 255) as u8,
            ),
        }
    }
}

#[test]
fn test_simple_blending() {
    // Opaque white over opaque black
    assert_eq!(
        BlendMode::Simple.blend(&Colour::WHITE, &Colour::BLACK),
        Colour::WHITE
    );

    // Opaque black over opaque white
    assert_eq!(
        BlendMode::Simple.blend(&Colour::BLACK, &Colour::WHITE),
        Colour::BLACK
    );

    // Transparent grey over opaque red
    assert_eq!(
        BlendMode::Simple.blend(&Colour::TRANSPARENT_GREY, &Colour::RED),
        Colour::rgba(191, 64, 64, 255)
    );

    // Transparent grey over transparent grey
    assert_eq!(
        BlendMode::Simple.blend(&Colour::TRANSPARENT_GREY, &Colour::TRANSPARENT_GREY),
        Colour::rgba(96, 96, 96, 191)
    );
}

#[test]
fn test_source_blending() {
    // Opaque white over opaque black
    assert_eq!(
        BlendMode::SourceAlpha.blend(&Colour::WHITE, &Colour::BLACK),
        Colour::WHITE
    );

    // Opaque black over opaque white
    assert_eq!(
        BlendMode::SourceAlpha.blend(&Colour::BLACK, &Colour::WHITE),
        Colour::BLACK
    );

    // Transparent grey over opaque red
    assert_eq!(
        BlendMode::SourceAlpha.blend(&Colour::TRANSPARENT_GREY, &Colour::RED),
        Colour::rgba(191, 64, 64, 255)
    );

    // Transparent grey over transparent grey
    assert_eq!(
        BlendMode::SourceAlpha.blend(&Colour::TRANSPARENT_GREY, &Colour::TRANSPARENT_GREY),
        Colour::rgba(128, 128, 128, 191)
    );
}
