use crate::{
    bitmap::{BitmapAccess, BitmapMutAccess},
    colour::{BlendMode, Colour},
    formats::{self, PixelFormatSpec},
    math::{PixelBox, PixelPoint},
};

mod dispatch;
use dispatch::{BinaryDispatch, GenericBinary, GenericUnary, UnaryDispatch, DISPATCH};

// operation implementations
mod composite;
mod copy_from;
mod fill;
mod fill_region;
mod fill_region_masked;
mod rectangle;

#[cfg(test)]
mod generate;

pub trait UnaryOpSpec: GenericUnary<Self::Params<'static>> {
    type Params<'l>;

    fn build<'l>() -> UnaryDispatch<Self::Params<'l>>;
}

pub trait BinaryOpSpec: GenericBinary<Self::Params<'static>> {
    type Params<'l>;

    fn build<'l>() -> BinaryDispatch<Self::Params<'l>>;
}

pub trait KahloOps<Format: PixelFormatSpec>: BitmapMutAccess<Format> {
    fn composite<SrcFormat: PixelFormatSpec>(
        &mut self,
        src: &dyn BitmapAccess<SrcFormat>,
        src_area: PixelBox,
        dst: PixelPoint,
        blend: BlendMode,
    ) {
        DISPATCH.composite.select::<Format, SrcFormat>()(
            self.data_mut(),
            src.data(),
            &(src_area, dst, blend),
        )
    }

    fn copy_from<SrcFormat: PixelFormatSpec>(
        &mut self,
        src: &dyn BitmapAccess<SrcFormat>,
        src_area: PixelBox,
        dst: PixelPoint,
    ) {
        DISPATCH.copy_from.select::<Format, SrcFormat>()(
            self.data_mut(),
            src.data(),
            &(src_area, dst),
        );
    }

    fn fill(&mut self, colour: Colour) {
        DISPATCH.fill.select::<Format>()(self.data_mut(), &(colour,));
    }

    fn fill_region(&mut self, region: PixelBox, colour: Colour) {
        DISPATCH.fill_region.select::<Format>()(self.data_mut(), &(region, colour));
    }

    fn fill_region_masked<'l>(
        &mut self,
        mask: &'l dyn BitmapAccess<formats::A8>,
        src_area: PixelBox,
        dst: PixelPoint,
        colour: Colour,
        blend: BlendMode,
    ) {
        DISPATCH
            .use_lifetime()
            .fill_region_masked
            .select::<Format>()(self.data_mut(), &(mask, src_area, dst, colour, blend))
    }

    fn rectangle(&mut self, area: PixelBox, width: usize, colour: Colour) {
        DISPATCH.rectangle.select::<Format>()(self.data_mut(), &(area, width, colour));
    }
}

impl<Format: PixelFormatSpec, T: BitmapMutAccess<Format>> KahloOps<Format> for T {}

#[cfg(test)]
mod fuzz;

#[cfg(test)]
mod benchmark;

#[cfg(test)]
pub fn collect_benchmarks(cri: &mut criterion::Criterion) {
    cri.benchmark_group("fill");
}
