use crate::{
    bitmap::{BitmapData, BitmapDataMut},
    colour::{BlendMode, Colour},
    formats::{PixelFormatSpec, A8},
    math::{clip_to, PixelBox, PixelPoint},
};

use super::{
    dispatch::{BinaryDispatch, GenericBinary},
    BinaryOpSpec,
};

pub struct CompositeOp;

impl<'l> GenericBinary<(PixelBox, PixelPoint, BlendMode)> for CompositeOp {
    fn perform<Format: PixelFormatSpec, SrcFormat: PixelFormatSpec>(
        write_data: &mut dyn BitmapDataMut,
        read_data: &dyn BitmapData,
        params: &(PixelBox, PixelPoint, BlendMode),
    ) {
        let (read_box, dst, blend) = params;

        let Some((write_box, read_box)) = clip_to(
            PixelBox::from_size(write_data.size()),
            PixelBox::from_size(read_data.size()),
            PixelBox::from_origin_and_size(*dst, read_data.size()),
            *read_box,
        ) else {
            return;
        };

        let read_stride = read_data.stride();
        let write_stride = write_data.stride();
        let read_data = read_data.data();
        let write_data = write_data.data_mut();
        for y in 0..read_box.height() {
            let y_offset_r = (y + read_box.min.y) as usize * read_stride;
            let y_offset_w = (y + write_box.min.y) as usize * write_stride;
            for x in 0..read_box.width() {
                let x_offset_r = (x + read_box.min.x) as usize * A8::PIXEL_WIDTH;
                let x_offset_w = (x + write_box.min.x) as usize * Format::PIXEL_WIDTH;

                let existing =
                    Colour::read_from_bytes::<Format>(&write_data[y_offset_w + x_offset_w..]);
                let source =
                    Colour::read_from_bytes::<SrcFormat>(&read_data[y_offset_r + x_offset_r..]);
                let new_colour = blend.blend(&source, &existing);
                new_colour.write_as_bytes::<Format>(&mut write_data[y_offset_w + x_offset_w..]);
            }
        }
    }
}

impl BinaryOpSpec for CompositeOp {
    type Params<'l> = (PixelBox, PixelPoint, BlendMode);
    fn build<'l>() -> BinaryDispatch<Self::Params<'l>> {
        BinaryDispatch::from_generic::<Self>()
    }
}
