use crate::{
    bitmap::{BitmapData, BitmapDataMut},
    colour::Colour,
    formats::{Abgr32, Bgr24, Bgr32, PixelFormatSpec, Rgb24, Rgb32, Rgba32, A8},
    math::{clip_to, PixelBox, PixelPoint},
};

use super::{dispatch::BinaryDispatch, BinaryOpSpec, GenericBinary};

pub struct CopyFromOp;

#[cfg(test)]
const _: () = {
    use super::generate;
    impl generate::BinaryInputSpec for CopyFromOp {
        fn gen<'l>() -> impl generate::ValueGeneratorList<'l, OutputTuple = Self::Params<'l>> {
            (generate::src_area, generate::dst_point)
        }
    }
};

impl BinaryOpSpec for CopyFromOp {
    type Params<'l> = (PixelBox, PixelPoint);

    fn build<'l>() -> BinaryDispatch<Self::Params<'l>> {
        BinaryDispatch::from_generic::<Self>()
            .with::<Bgr32, Rgba32>(copy_bgr32_rgba32)
            .with::<A8, A8>(copy_same_fmt::<{ A8::PIXEL_WIDTH }>)
            .with::<Abgr32, Abgr32>(copy_same_fmt::<{ Abgr32::PIXEL_WIDTH }>)
            .with::<Bgr24, Bgr24>(copy_same_fmt::<{ Bgr24::PIXEL_WIDTH }>)
            .with::<Bgr32, Bgr32>(copy_same_fmt::<{ Bgr32::PIXEL_WIDTH }>)
            .with::<Rgb24, Rgb24>(copy_same_fmt::<{ Rgb24::PIXEL_WIDTH }>)
            .with::<Rgb32, Rgb32>(copy_same_fmt::<{ Rgb32::PIXEL_WIDTH }>)
            .with::<Rgba32, Rgba32>(copy_same_fmt::<{ Rgba32::PIXEL_WIDTH }>)
    }
}

/// read rgba32 and write bgr32
fn copy_bgr32_rgba32(
    write_data: &mut dyn BitmapDataMut,
    read_data: &dyn BitmapData,
    params: &(PixelBox, PixelPoint),
) {
    let read_box = params.0;
    let write_box = PixelBox::from_origin_and_size(params.1, read_box.size());
    let Some((read_box, write_box)) = clip_to(
        PixelBox::from_size(read_data.size()),
        PixelBox::from_size(write_data.size()),
        read_box,
        write_box,
    ) else {
        // nothing to do
        return;
    };

    let write_stride = write_data.stride();
    let write_data = write_data.data_mut();
    let read_stride = read_data.stride();
    let read_data = read_data.data();

    fn do_copy(target: &mut [u8], source: &[u8], count: usize) {
        // for now, do the dumb version
        let source = unsafe { std::slice::from_raw_parts(source.as_ptr() as *const u32, count) };
        let target = unsafe { std::slice::from_raw_parts_mut(target.as_ptr() as *mut u32, count) };
        for i in 0..count {
            target[i] = source[i].to_be() >> 8;
        }
    }

    for y in 0..read_box.height() {
        let y_offset_r = (y + read_box.min.y) as usize * read_stride;
        let y_offset_w = (y + write_box.min.y) as usize * write_stride;
        do_copy(
            &mut write_data[(y_offset_w + write_box.min.x as usize * Bgr32::PIXEL_WIDTH)..],
            &read_data[(y_offset_r + read_box.min.x as usize * Rgba32::PIXEL_WIDTH)..],
            read_box.width() as usize,
        );
    }
}

fn copy_same_fmt<const WIDTH: usize>(
    write_data: &mut dyn BitmapDataMut,
    read_data: &dyn BitmapData,
    params: &(PixelBox, PixelPoint),
) {
    let read_box = params.0;
    let write_box = PixelBox::from_origin_and_size(params.1, read_box.size());
    let Some((read_box, write_box)) = clip_to(
        PixelBox::from_size(read_data.size()),
        PixelBox::from_size(write_data.size()),
        read_box,
        write_box,
    ) else {
        // nothing to do
        return;
    };

    let write_stride = write_data.stride();
    let write_data = write_data.data_mut();
    let read_stride = read_data.stride();
    let read_data = read_data.data();

    let read_line_offset = read_box.min.x as usize * WIDTH;
    let write_line_offset = write_box.min.x as usize * WIDTH;
    let line_width = read_box.width() as usize * WIDTH;

    for y in 0..read_box.height() {
        let write_slice =
            &mut write_data[(y + write_box.min.y) as usize * write_stride + write_line_offset..];
        let read_slice =
            &read_data[(y + read_box.min.y) as usize * read_stride + read_line_offset..];
        (&mut write_slice[..line_width]).copy_from_slice(&read_slice[..line_width]);
    }
}

impl GenericBinary<(PixelBox, PixelPoint)> for CopyFromOp {
    fn perform<Format: PixelFormatSpec, ReadFormat: PixelFormatSpec>(
        write_data: &mut dyn BitmapDataMut,
        read_data: &dyn BitmapData,
        params: &(PixelBox, PixelPoint),
    ) {
        let read_box = params.0;
        let write_box = PixelBox::from_origin_and_size(params.1, read_box.size());
        let Some((read_box, write_box)) = clip_to(
            PixelBox::from_size(read_data.size()),
            PixelBox::from_size(write_data.size()),
            read_box,
            write_box,
        ) else {
            // nothing to do
            return;
        };

        let write_stride = write_data.stride();
        let write_data = write_data.data_mut();
        let read_stride = read_data.stride();
        let read_data = read_data.data();

        for y in 0..read_box.height() {
            let write_slice = &mut write_data[(y + write_box.min.y) as usize * write_stride..];
            let read_slice = &read_data[(y + read_box.min.y) as usize * read_stride..];
            for x in 0..read_box.width() {
                Colour::convert_byte_form::<ReadFormat, Format>(
                    &mut write_slice[(x + write_box.min.x) as usize * Format::PIXEL_WIDTH..],
                    &read_slice[(x + read_box.min.x) as usize * ReadFormat::PIXEL_WIDTH..],
                );
            }
        }
    }
}
