use crate::{
    bitmap::{BitmapAccess, BitmapDataMut},
    colour::{BlendMode, Colour},
    formats::{PixelFormatSpec, A8},
    math::{clip_to, PixelBox, PixelPoint},
};

use super::{
    dispatch::{GenericUnary, UnaryDispatch},
    UnaryOpSpec,
};

pub struct FillRegionMaskedOp;

type OpParams<'l> = (
    &'l dyn BitmapAccess<A8>,
    PixelBox,
    PixelPoint,
    Colour,
    BlendMode,
);

#[cfg(test)]
const _: () = {
    use super::generate;
    impl generate::UnaryInputSpec for FillRegionMaskedOp {
        fn gen<'l>() -> impl generate::ValueGeneratorList<'l, OutputTuple = Self::Params<'l>> {
            (
                generate::alphamap,
                generate::alphamap_area,
                generate::dst_point,
                generate::random_colour,
                generate::constant(BlendMode::Simple),
            )
        }
    }
};

impl<'l> GenericUnary<OpParams<'l>> for FillRegionMaskedOp {
    fn perform<Format: PixelFormatSpec>(write_data: &mut dyn BitmapDataMut, params: &OpParams<'l>) {
        let (read_alphamap, read_box, dst, colour, blend) = params;

        let Some((write_box, read_box)) = clip_to(
            PixelBox::from_size(write_data.size()),
            PixelBox::from_size(read_alphamap.size()),
            PixelBox::from_origin_and_size(*dst, read_alphamap.size()),
            *read_box,
        ) else {
            return;
        };

        let read_stride = read_alphamap.row_stride();
        let write_stride = write_data.stride();
        let read_data = read_alphamap.data().data();
        let write_data = write_data.data_mut();
        for y in 0..read_box.height() {
            let y_offset_r = (y + read_box.min.y) as usize * read_stride;
            let y_offset_w = (y + write_box.min.y) as usize * write_stride;
            for x in 0..read_box.width() {
                let x_offset_r = (x + read_box.min.x) as usize * A8::PIXEL_WIDTH;
                let x_offset_w = (x + write_box.min.x) as usize * Format::PIXEL_WIDTH;

                let existing =
                    Colour::read_from_bytes::<Format>(&write_data[y_offset_w + x_offset_w..]);
                let alpha_override = read_data[y_offset_r + x_offset_r];
                let new_colour = blend.blend(&colour.with_a(alpha_override), &existing);
                new_colour.write_as_bytes::<Format>(&mut write_data[y_offset_w + x_offset_w..]);
            }
        }
    }
}

impl UnaryOpSpec for FillRegionMaskedOp {
    type Params<'l> = OpParams<'l>;
    fn build<'l>() -> UnaryDispatch<Self::Params<'l>> {
        UnaryDispatch::from_generic::<Self>()
    }
}
