use rand::{
    distributions::{uniform::SampleUniform, Distribution, Standard, Uniform},
    prelude::*,
    rngs::ThreadRng,
    thread_rng, Fill,
};
use std::ops::Range;

use crate::{
    colour::Colour,
    formats::{self, PixelFormatSpec},
    math::{PixelBox, PixelPoint, PixelSize},
    prelude::*,
    Bitmap,
};

use super::{Context, ContextValueTag};

const DEFAULT_SIZE_BOUNDS: Range<i32> = 4..64;
pub struct ImageSizeBounds;
impl ContextValueTag for ImageSizeBounds {
    type Value = Range<i32>;
}

pub struct DstSize;
impl ContextValueTag for DstSize {
    type Value = PixelSize;
}

pub struct SrcSize;
impl ContextValueTag for SrcSize {
    type Value = PixelSize;
}

fn point_in_size(rng: &mut ThreadRng, sz: &PixelSize) -> PixelPoint {
    let xdist = Uniform::new(0, sz.width);
    let ydist = Uniform::new(0, sz.height);
    PixelPoint::new(xdist.sample(rng), ydist.sample(rng))
}

pub fn constant<V: Clone>(v: V) -> impl Fn(&mut Context) -> Option<V> {
    move |_| Some(v.clone())
}

pub fn uniform<V: SampleUniform + Clone>(from: V, to: V) -> impl Fn(&mut Context) -> Option<V> {
    move |ctx| {
        let uniform = Uniform::new(from.clone(), to.clone());
        Some(ctx.rng().sample(uniform))
    }
}

pub fn random_colour(ctx: &mut Context) -> Option<Colour> {
    let rng = ctx.rng();
    Some(Colour::rgba(rng.gen(), rng.gen(), rng.gen(), rng.gen()))
}

pub fn dst_image_generator<Format: PixelFormatSpec>(ctx: &mut Context) -> Option<Bitmap<Format>> {
    let size_bounds = ctx
        .lookup::<ImageSizeBounds>()
        .unwrap_or(&DEFAULT_SIZE_BOUNDS);
    let size_dist = Uniform::new(size_bounds.start, size_bounds.end);
    let mut bitmap = Bitmap::new(
        size_dist.sample(ctx.rng()) as usize,
        size_dist.sample(ctx.rng()) as usize,
    );
    bitmap.data_mut().data_mut().try_fill(ctx.rng()).ok()?;
    ctx.insert::<DstSize>(bitmap.size());
    Some(bitmap)
}

pub fn src_image_generator<Format: PixelFormatSpec>(ctx: &mut Context) -> Option<Bitmap<Format>> {
    let size_bounds = ctx
        .lookup::<ImageSizeBounds>()
        .unwrap_or(&DEFAULT_SIZE_BOUNDS);
    let size_dist = Uniform::new(size_bounds.start, size_bounds.end);
    let mut bitmap = Bitmap::new(
        size_dist.sample(ctx.rng()) as usize,
        size_dist.sample(ctx.rng()) as usize,
    );
    bitmap.data_mut().data_mut().try_fill(ctx.rng()).ok()?;
    ctx.insert::<SrcSize>(bitmap.size());
    Some(bitmap)
}

pub fn alphamap<'l>(ctx: &mut Context) -> Option<&'l dyn BitmapAccess<formats::A8>> {
    None
}

pub fn alphamap_area(ctx: &mut Context) -> Option<PixelBox> {
    None
}

pub fn dst_point(ctx: &mut Context) -> Option<PixelPoint> {
    let mut rng = thread_rng();
    Some(point_in_size(&mut rng, ctx.lookup::<DstSize>()?))
}

pub fn dst_area(ctx: &mut Context) -> Option<PixelBox> {
    let mut rng = thread_rng();
    let src_size = ctx.lookup::<DstSize>()?;
    Some(PixelBox::new(
        point_in_size(&mut rng, src_size),
        point_in_size(&mut rng, src_size),
    ))
}

pub fn src_area(ctx: &mut Context) -> Option<PixelBox> {
    let mut rng = thread_rng();
    let src_size = ctx.lookup::<SrcSize>()?;
    Some(PixelBox::new(
        point_in_size(&mut rng, src_size),
        point_in_size(&mut rng, src_size),
    ))
}
